/*
 * Scratchpad MMU - Simulator main module
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <stdint.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>


#include "mb_mem.h"
#include "mb_core.h"
#include "mb_elf.h"
#include "settings.h"

typedef struct Carousel_Entry_struct {
        unsigned        base, scratch_flag;
        unsigned        page[BLOCK_SIZE / 4];
} Carousel_Entry;

#define METRIC_TIME         0
#define METRIC_SYS_TIME     1
#define METRIC_MISS         2
#define METRIC_OPEN_CLOSE   3
#define METRIC_PREEMPTS     4
#define NUM_METRICS         5
static const char * metric_name[] = {
            "time", "syscs", "miss", "open/close", 
            "preempts", 0 };

typedef struct Metric_struct {
        uint64_t        best, worst, total;
} Metric;

typedef struct Active_Task_struct {
        unsigned        function;
        uint64_t        metric[NUM_METRICS];
        const char *    name;
} Active_Task;

#define SPECIAL_IDLE        0
#define SPECIAL_BOOT        1
#define SPECIAL_SCHED_TO    2
#define SPECIAL_SCHED_FROM  3
#define MAX_KNOWN_TASKS     (MAX_THREADS + 4)

#define MAX_NAME_LEN        8

typedef struct Known_Task_struct {
        unsigned        function;
        unsigned        start_count, complete_count;
        uint64_t        total_time;
        Metric          metric[NUM_METRICS];
        char            name[MAX_NAME_LEN + 1];
} Known_Task;

typedef enum { DUE_TO_EXIT, DUE_TO_INTERRUPT,
        DUE_TO_END, HAMMER_TIME } Stop_Type;

static uint64_t abs_time;
static uint64_t trigger_time;
static Stop_Type stop = HAMMER_TIME;
static MB_Context * mb ;
static MB_Context mb_orig ;
static FILE * trace ;
static Carousel_Entry carousel[NUM_BLOCKS];
static unsigned fsl_source_reg;
static unsigned fsl_dest_reg;
static unsigned scheduler_next_task, scheduler_end_task;
static unsigned num_active_tasks, interrupt_handler;
static unsigned num_known_tasks;
static uint64_t run_budget, boot_time;
static Active_Task active_task[MAX_THREADS];
static Known_Task known_task[MAX_KNOWN_TASKS];
static uint64_t user_metric[NUM_METRICS];
static uint64_t sys_metric[NUM_METRICS];
static uint64_t next_quantum, quanta = 0;
static unsigned store_queue;
static unsigned dcache_writeback = 0;
static unsigned dcache_hash = 0;
static const unsigned system_spm = 1;
static unsigned icache_store[ICACHE_STORE_SIZE / 4];
static unsigned icache_tags[ICACHE_STORE_SIZE / ICACHE_LINE_SIZE];
static unsigned dcache_store[DCACHE_STORE_SIZE / 4];
static unsigned dcache_tags[DCACHE_STORE_SIZE / DCACHE_LINE_SIZE];
static unsigned char dcache_dirty[DCACHE_STORE_SIZE / 4];
static uint64_t last_ie_time, longest_cs_time;
    
static unsigned Hash_Func(unsigned address)
{
    address /= DCACHE_LINE_SIZE;
    if (dcache_hash) {
        address ^= (address >> 6);
    }
    
    return address % (DCACHE_STORE_SIZE / DCACHE_LINE_SIZE);
}

static unsigned Transfer_Cost(unsigned bytes)
{
    unsigned full_transactions = (bytes / EXT_MEM_TRANS_BYTES);
    unsigned cost = full_transactions *
            (EXT_MEM_LATENCY + (EXT_MEM_TRANS_BYTES / EXT_MEM_BUS_WIDTH));
   
    assert(bytes);
    bytes -= full_transactions * EXT_MEM_TRANS_BYTES;
    if (bytes) {
        cost += EXT_MEM_LATENCY + 
                ((bytes + EXT_MEM_BUS_WIDTH - 1) / EXT_MEM_BUS_WIDTH);
    }
    return cost;
}

static unsigned Get_Mode(void)
{
    unsigned mode = (mb->msr & MSR_MASK);

    if (trigger_time == 0) {
        return MODE_SYSTEM;
    }
    assert ((mode == MODE_USER) || (mode == MODE_SYSCALL)
            || (mode == MODE_SYSTEM));
    return mode;
}

static unsigned In_User(void)
{
    return (Get_Mode() != MODE_SYSTEM);
}

static void Trace ( void * t_user , MB_Context * mc , 
                            MB_Trace_Name trace_name , const void * param )
{
    unsigned param_u = (unsigned) ( (uintptr_t) param ) ;

    if ((trace_name == MB_SIM_CMD) && (param_u == 1)) {
        stop = DUE_TO_EXIT;
    }
    if (trace_name == MB_INTERRUPT) {
        fprintf(trace, "interrupt:\n");
    }
    if (trace_name == MB_ILLEGAL_INST) {
        fprintf(stderr, "illegal instruction %x at %x\n",
                    mc->cur_iword, mc->cur_pc);
        exit(1);
    }
    if (trace_name == MB_EXECUTE) {
        fprintf(trace, "%010ju %08x %08x %s" , 
                    abs_time, mc -> cur_pc, mc -> cur_iword,
                    (const char *) param);
    }
}

static void Metric_Add(unsigned nr, uint64_t add)
{
    if (In_User()) {
        user_metric[nr] += add;
    } else {
        sys_metric[nr] += add;
    }
}

static void Time_Add(uint64_t add, const char * activity)
{
    abs_time += add;
    if (Get_Mode() == MODE_SYSCALL) {
        Metric_Add(METRIC_SYS_TIME, add);
    } else {
        Metric_Add(METRIC_TIME, add);
    }
    if (add > 10) {
        fprintf(trace, "costly %s %ju\n", activity, add);
    }

    while ((next_quantum != 0) && (next_quantum <= abs_time)) {
        if (In_User()) {
            assert(num_active_tasks);
            fprintf(trace, "quantum %ju %s %s %x\n", quanta,
                    active_task[num_active_tasks - 1].name,
                    activity, mb->cur_pc);
        } else {
            fprintf(trace, "quantum %ju sys %s %x\n", 
                    quanta, activity, mb->cur_pc);
        }
        quanta ++;
        next_quantum += TIMER_QUANTUM;
    }
}

static const char * Lookup_Name(unsigned function)
{
    unsigned i;
    for (i = 0; i < num_known_tasks; i++) {
        if (known_task[i].function == function) {
            return known_task[i].name;
        }
    }
    return 0;
}

static Known_Task * Account_Task(unsigned function, 
            uint64_t * metric, int completed)
{
    Known_Task * t = 0;
    unsigned i;

    for (i = 0; i < num_known_tasks; i++) {
        if (known_task[i].function == function) {
            t = &known_task[i];
            break;
        }
    }
    if (!t) {
        assert(num_known_tasks < MAX_KNOWN_TASKS);
        t = &known_task[num_known_tasks];
        t->function = function;
        snprintf(t->name, MAX_NAME_LEN, "%08x", known_task[i].function);
        t->name[MAX_NAME_LEN] = '\0';
        num_known_tasks ++;
    }
    if (!metric) {
        return t;
    }

    t->start_count++;
    t->total_time += metric[METRIC_TIME] + metric[METRIC_SYS_TIME];

    if (completed) {
        t->complete_count ++;

        for (i = 0; i < NUM_METRICS; i++) {
            t->metric[i].total += metric[i];

            if ((t->metric[i].best == 0)
            || (metric[i] < t->metric[i].best)) {
                t->metric[i].best = metric[i];
            }
            if (metric[i] > t->metric[i].worst) {
                t->metric[i].worst = metric[i];
            }
        }
    }
    for (i = 0; i < NUM_METRICS; i++) {
        metric[i] = 0;
    }
    return t;
}


static Carousel_Entry * Search(void * m_user, unsigned address)
{
    unsigned i, j;

    if (mb->msr & (MSR_DCACHE_ENABLE|MSR_ICACHE_ENABLE)) {
        return NULL;
    }
    address &= ~(BLOCK_SIZE - 1);
    if (address == 0) {
        return NULL;
    }
    for (i = 0; i < NUM_BLOCKS; i++) {
        if ((carousel[i].base == address) && (!carousel[i].scratch_flag)) {
            return &carousel[i];
        }
    }
    for (i = 0; i < NUM_BLOCKS; i++) {
        if (carousel[i].base == address) {
            return &carousel[i];
        }
    }
    return NULL;
}

static void Flush_Store_Queue(void)
{
#ifdef STORE_QUEUE
    if (store_queue) {
        mb->clock_cycle += store_queue;
        store_queue = 0;
    }
#endif
}

static unsigned Service_Writeback(void * m_user, unsigned cache_line)
{
    int wb = 0;
    unsigned a1 = (dcache_tags[cache_line] * DCACHE_LINE_SIZE);
    unsigned i;

    for (i = 0; i < DCACHE_LINE_SIZE; i++) {
        if (dcache_dirty[((a1 % DCACHE_STORE_SIZE) + i) / 4]) {
            unsigned byte = Page_Load(dcache_store,
                (cache_line * DCACHE_LINE_SIZE) + i, 1);
            mb_orig.store_fn(m_user, a1 + i, byte, 1);
            wb = 1;
            dcache_dirty[((a1 % DCACHE_STORE_SIZE) + i) / 4] = 0;
        }
    }
    if (wb) {
        return Transfer_Cost(DCACHE_LINE_SIZE);
    }
    return 0;
}

static unsigned Service_Data_Miss(void * m_user, unsigned address)
{
    unsigned cache_line = Hash_Func(address);
    unsigned a0 = (address / DCACHE_LINE_SIZE) * DCACHE_LINE_SIZE;
    unsigned i;
    unsigned timing = Transfer_Cost(DCACHE_LINE_SIZE);

    Metric_Add(METRIC_MISS, 1);
    if (dcache_writeback) {
        timing += Service_Writeback(m_user, cache_line);
    }
    for (i = 0; i < DCACHE_LINE_SIZE; i++) {
        unsigned byte = mb_orig.load_fn(m_user, a0 + i, 1);
        Page_Store(dcache_store,
            (cache_line * DCACHE_LINE_SIZE) + i,
            byte, 1);
    }
    if (boot_time != 0) {
        fprintf(trace, "data miss\n");
    }

    dcache_tags[cache_line] = address / DCACHE_LINE_SIZE;
    return timing;
}

static void C_Store (void * m_user, unsigned address, 
                    unsigned data, unsigned size)
{
    Carousel_Entry * ce = Search(m_user, address);

    Flush_Store_Queue();
    if (!ce) {
        unsigned clock_cycle = mb->clock_cycle;
        mb_orig.store_fn(m_user, address, data, size);

        if (system_spm
        && ((address < SPM_LIMIT) || (address >= USER_LIMIT))) {
            /* system SPM */
            mb->clock_cycle = clock_cycle;
        } else {
            int write_through = 1;
            int hit = 0;
            unsigned cache_line = Hash_Func(address);

            if (boot_time == 0) {
                fprintf(trace, "STORE %x\n", address);
            } 
            if ((mb->msr & MSR_DCACHE_ENABLE)
            && (address < USER_LIMIT)
            && (dcache_tags[cache_line] == (address / DCACHE_LINE_SIZE))) {

                /* the line is in cache - update cache */
                hit = 1;
                if (dcache_writeback) {
                    write_through = 0;
                    mb->clock_cycle = clock_cycle;
                }

            } else if (dcache_writeback 
            && (mb->msr & MSR_DCACHE_ENABLE)
            && (address < USER_LIMIT)) {

                /* Line is not in cache but can be brought there */
                mb->clock_cycle = clock_cycle + 
                        Service_Data_Miss(m_user, address);
                write_through = 0;
                hit = 1;

            } 

            if (hit) {
                Page_Store(dcache_store,
                    (cache_line * DCACHE_LINE_SIZE) +
                    (address % DCACHE_LINE_SIZE), data, size);
                dcache_dirty[(address % DCACHE_STORE_SIZE) / 4] = 1;
            }
            if (write_through) {
#ifdef STORE_QUEUE
                store_queue = mb->clock_cycle - clock_cycle;
                mb->clock_cycle = clock_cycle;
#endif
                Metric_Add(METRIC_MISS, 1);
            }
        }
        return;
    }
    address -= ce->base;
    Page_Store(ce->page, address, data, size);
}

static unsigned Generic_Load(void * m_user, unsigned address, 
                    unsigned size, int data_bus)
{
    Carousel_Entry * ce = Search(m_user, address);

    if (data_bus) {
        Flush_Store_Queue();
    }
    if (!ce) {
        unsigned clock_cycle = mb->clock_cycle;
        unsigned d = mb_orig.load_fn(m_user, address, size);
        unsigned cache_line = Hash_Func(address);

        if (system_spm
        && ((address < SPM_LIMIT) || (address >= USER_LIMIT))) {
            /* system SPM  */
            mb->clock_cycle = clock_cycle;

        } else if (data_bus 
        && (mb->msr & MSR_DCACHE_ENABLE)
        && (address < USER_LIMIT)
        && (dcache_tags[cache_line] == (address / DCACHE_LINE_SIZE))) {
            /* data cache hit */
            mb->clock_cycle = clock_cycle;
            return Page_Load(dcache_store, 
                    (cache_line * DCACHE_LINE_SIZE) +
                    (address % DCACHE_LINE_SIZE), size);

        } else if ((!data_bus)
        && (mb->msr & MSR_ICACHE_ENABLE)
        && (address < USER_LIMIT)
        && (icache_tags[ (address / ICACHE_LINE_SIZE) % 
                (ICACHE_STORE_SIZE / ICACHE_LINE_SIZE) ] ==
                    (address / ICACHE_STORE_SIZE))) {
            /* instruction cache hit */
            mb->clock_cycle = clock_cycle;
            return Page_Load(icache_store, 
                    address % ICACHE_STORE_SIZE, size);

    /**** After this point, it must be a miss ****/
        } else if (data_bus 
        && (mb->msr & MSR_DCACHE_ENABLE)
        && (address < USER_LIMIT)) {
            /* data cache miss */
            mb->clock_cycle = clock_cycle + 
                Service_Data_Miss(m_user, address);

        } else if ((!data_bus)
        && (mb->msr & MSR_ICACHE_ENABLE)
        && (address < USER_LIMIT)) {
            /* instruction cache miss */
            unsigned i, a0 = (address / ICACHE_LINE_SIZE) * ICACHE_LINE_SIZE;
            Metric_Add(METRIC_MISS, 1);
            for (i = 0; i < ICACHE_LINE_SIZE; i++) {
                unsigned byte = mb_orig.load_fn(m_user, a0 + i, 1);
                Page_Store(icache_store,
                    (a0 % ICACHE_STORE_SIZE) + i, byte, 1);
            }
            icache_tags[ (address / ICACHE_LINE_SIZE) % 
                (ICACHE_STORE_SIZE / ICACHE_LINE_SIZE) ] =
                    address / ICACHE_STORE_SIZE;
            mb->clock_cycle = clock_cycle + Transfer_Cost(ICACHE_LINE_SIZE);

            if (boot_time != 0) {
                fprintf(trace, "instruction miss\n");
            }

        } else {
            /* no-cache miss */
            Metric_Add(METRIC_MISS, 1);

            if (boot_time != 0) {
                fprintf(trace, "%s miss t1\n",
                        data_bus ? "data" : "instruction" );
            } else {
                fprintf(trace, "LOAD %x\n", address);
            }
        }
        return d;
    }
    address -= ce->base;
    return Page_Load(ce->page, address, size);
}

static unsigned C_Load ( void * m_user , unsigned address , unsigned size )
{
    return Generic_Load(m_user, address, size, 1);
}

static unsigned C_IFetch ( void * m_user , unsigned address , unsigned size )
{
    return Generic_Load(m_user, address, size, 0);
}


static unsigned Basereg_Load(void * io_user, unsigned address, unsigned size)
{
    address &= (1 << PAGE_SHIFT) - 1;
    address /= 4;
    if (address >= NUM_BLOCKS) {
        fprintf(stderr, "Basereg_Load: No basereg numbered %u\n", address);
        exit(1);
    }
    return carousel[address].base;
}

static void Basereg_Store ( void * io_user , unsigned address , 
                            unsigned data , unsigned size )

{
    fprintf(stderr, "Must not write to basereg area\n");
    exit(1);
}

static unsigned Timer_Load ( void * io_user , unsigned address , unsigned size )
{
    return (unsigned) (abs_time / (uint64_t) TIMER_QUANTUM);
}

static void Timer_Store ( void * io_user , unsigned address , 
                            unsigned data , unsigned size )
{
    if (trigger_time == 0) {
        fprintf(trace, "BOOT time %ju\n", sys_metric[METRIC_TIME]);
        Account_Task(SPECIAL_BOOT, sys_metric, 1);
        memset(sys_metric, 0, sizeof(sys_metric));
        boot_time = abs_time;
        next_quantum = abs_time;
    }
    if (data == 0) {
        trigger_time = 1;
    } else {
        trigger_time = ((uint64_t) data) * TIMER_QUANTUM;
    }
    fprintf(trace, "next interrupt at %ju\n", trigger_time);
}

static void C_Report(void)
{
    unsigned i;
    fprintf(trace, "Carousel:");
    for (i = 0; i < NUM_BLOCKS; i++) {
        fprintf(trace, " %08x", carousel[i].base);
    }
    fprintf(trace, "\n");
}

static void Program_Aborted ( int signum, siginfo_t * siginfo, void * data )
{
    if (mb != NULL) {
        char buf[BUFSIZ];
        snprintf(buf, BUFSIZ - 1,
                "Aborted at PC %08x, instruction %08x, r15 = %08x\n" , 
                mb -> cur_pc , mb -> cur_iword,
                mb -> gpr [ 15 ] ) ;
        buf[BUFSIZ - 1] = '\0';
        fputs(buf, stderr);
        fputs(buf, trace);
        C_Report();
    }
    exit ( 1 );
}

static void Clean_Stop ( int signum, siginfo_t * siginfo, void * data )
{
    stop = DUE_TO_INTERRUPT;
}

static void C_Put(void * m_user, unsigned flags, unsigned fsl, unsigned data)
{
    unsigned clock_cycle = mb->clock_cycle;
    unsigned i, size, byte, is_open, cr;
    uint64_t miss_count = In_User() ? 
            user_metric[METRIC_MISS] : sys_metric[METRIC_MISS];
    
    fsl = fsl << 4;
    if (flags & MB_FSL_C) {
        fsl |= 1;
    } else if (flags & MB_FSL_N) {
        fsl |= 2;
    }
    switch(fsl) {
    case 0x00:
        /* put fsl0 */
        fsl_source_reg = data;
        break;
    case 0x10:
        /* put fsl1 */
        fsl_dest_reg = data;
        break;
    case 0x01:
        /* cput fsl0 dma transfer */
        size = data;
        Time_Add(Transfer_Cost(size), "dma");
        while (size) {
            byte = mb_orig.load_fn(m_user, fsl_source_reg, 1);
            mb_orig.store_fn(m_user, fsl_dest_reg, byte, 1);
            fprintf(trace, "dmacpy: %08x <- %08x = %02x\n", 
                    fsl_dest_reg, fsl_source_reg, byte);
            fsl_source_reg++;
            fsl_dest_reg++;
            size--;
        }
        break;
    case 0x11:
        /* cput fsl1 OPEN/CLOSE */
        size = data & OPEN_SIZE_BITS;
        if (size == 0) {
            size = 1;
        }
        if (mb->msr & (MSR_DCACHE_ENABLE|MSR_ICACHE_ENABLE)) {
            fprintf(stderr, "OPEN/CLOSE: refusing to allow Carousel "
                    "while IC/DC active.\n");
            exit(1);
        }
        is_open = (fsl_source_reg != 0xffffffff);
        if (fsl_dest_reg >= NUM_BLOCKS) {
            fprintf(stderr, "OPEN/CLOSE: No block numbered %u\n", 
                    fsl_dest_reg);
            exit(1);
        }
        while (size) {
            if (data & OPEN_DISCARD_FLAG) {
                /* discard CLOSE operation (no copy) */
                assert(!is_open);
                fprintf(trace, "discard %u\n", fsl_dest_reg);
                carousel[fsl_dest_reg].base = 0;

            } else if (is_open) {
                /* various sorts of OPEN */
                if (carousel[fsl_dest_reg].base) {
                    fprintf(stderr, "Double OPEN on block %u\n",
                            fsl_dest_reg);
                    exit(1);
                }
                if (fsl_source_reg & (BLOCK_SIZE - 1)) {
                    fprintf(stderr, "Unaligned OPEN %x on block %u\n",
                            fsl_source_reg, fsl_dest_reg);
                    exit(1);
                }
                if (data & OPEN_SCRATCH_FLAG) {
                    fprintf(trace, "scratch open %x to %u\n", 
                            fsl_source_reg, fsl_dest_reg);
                    carousel[fsl_dest_reg].scratch_flag = 1;
                } else {
                    Carousel_Entry * previous = Search(m_user, fsl_source_reg);
                    fprintf(trace, "normal open %x to %u\n", 
                            fsl_source_reg, fsl_dest_reg);

                    carousel[fsl_dest_reg].scratch_flag = 0;
                    Time_Add(Transfer_Cost(BLOCK_SIZE), "open");
                    byte = 0xff;
                    for (i = 0; i < BLOCK_SIZE; i++) {
                        if (fsl_source_reg) {
                            if (previous) {
                                byte = Page_Load(previous->page, i, 1);
                            } else {
                                byte = mb_orig.load_fn(m_user, 
                                        fsl_source_reg + i, 1);
                            }
                        }
                        Page_Store(carousel[fsl_dest_reg].page, i, byte, 1);
                        fprintf(trace, "open: %08x <- %08x = %02x\n", 
                                fsl_dest_reg + i, fsl_source_reg + i, byte);
                    }
                }
                carousel[fsl_dest_reg].base = fsl_source_reg;
                fsl_source_reg += BLOCK_SIZE;
                Metric_Add(METRIC_OPEN_CLOSE, 1);

            } else {
                /* normal CLOSE with copy */
                unsigned out_reg = carousel[fsl_dest_reg].base;
                Carousel_Entry * src = 0;
                Carousel_Entry * next = 0;
                
                fprintf(trace, "close %x to %u\n", 
                       out_reg, fsl_dest_reg);
                /* if (!out_reg) {
                    fprintf(stderr, "Empty CLOSE on block %u\n",
                            fsl_dest_reg);
                    exit(1);
                } */
                src = Search(m_user, out_reg);
                if (!src) {
                    /* This happens if out_reg == 0 */
                    assert(out_reg == 0);
                    src = &carousel[fsl_dest_reg];
                }
                carousel[fsl_dest_reg].base = 0;
                next = Search(m_user, out_reg);
                Time_Add(Transfer_Cost(BLOCK_SIZE), "close");
                for (i = 0; i < BLOCK_SIZE; i++) {
                    byte = Page_Load(src->page, i, 1);
                    if (out_reg) {
                        if (next) {
                            Page_Store(next->page, i, byte, 1);
                        }
                        mb_orig.store_fn(m_user, out_reg + i, byte, 1);
                        fprintf(trace, "close: %08x <- %08x = %02x\n", 
                                out_reg + i, fsl_dest_reg + i, byte);
                    }
                }
                Metric_Add(METRIC_OPEN_CLOSE, 1);
            }

            fsl_dest_reg = (fsl_dest_reg + 1) % NUM_BLOCKS;
            size --;
        }
        C_Report();
        break;
    case 0x21: 
        /* cput fsl2 print string */
        fprintf(trace, "print: ");
        byte = mb_orig.load_fn(m_user, data, 1);
        size = 0;
        cr = 0;
        while (byte) {
            if (!(isprint(byte) || isspace(byte))) {
                fprintf(stderr, "attempted to print unprintable "
                    "byte %d from %x\n",
                        byte, data);
                exit(1);
            }
            size ++;
            if (size > 1000) {
                fprintf(stderr, "runaway string!\n");
                exit(1);
            }
            fprintf(trace, "%c", byte);
            printf("%c", byte);
            cr = (byte == '\n');
            data ++;
            byte = mb_orig.load_fn(m_user, data, 1);
        } 
        if (!cr) {
            fprintf(trace, "\n");
            printf("\n");
        }
        fflush(stdout);
        break;
    case 0x31:
        /* cput fsl3 halt */
        if (trigger_time == 0) {
            fprintf(stderr, "Halt with unprogrammed timer\n");
            exit(1);
        } else if ((!In_User()) || !(mb->msr & MSR_IE)) {
            /* won't halt with interrupts disabled */
            fprintf(trace, "HALT skipped (MSR_IE)\n");

        } else if ((!In_User()) || (num_active_tasks != 1)) {
            fprintf(trace, "HALT skipped (RETI)\n");

        } else if (abs_time > trigger_time) {
            /* This can happen if the idle task resumes */
            fprintf(trace, "HALT skipped (too late)\n");

        } else {
            uint64_t halt_time = trigger_time - abs_time;
            fprintf(trace, "HALT time %ju\n", halt_time);
            Time_Add(halt_time, "halt");
        }
        break;
    case 0x41:
        /* cput fsl4 debug / capture */
        switch (data) {
        case TRC_CAPT_CALL_FUNCTION:
            fprintf(trace, "CALL %s pc %x r1 %x\n", 
                    Lookup_Name(fsl_source_reg),
                    fsl_source_reg, fsl_dest_reg);
            break;
        case TRC_CAPT_RETURN_FUNCTION:
            fprintf(trace, "RETURN %s max %x r1 %x\n", 
                    Lookup_Name(fsl_source_reg),
                    fsl_source_reg, fsl_dest_reg);
            break;
        case TRC_TASK_START:
            fprintf(trace, "TASK_START %s pc %x\n", 
                    Lookup_Name(fsl_source_reg), fsl_source_reg);
            assert((scheduler_next_task == 0)
                    || (scheduler_next_task == fsl_source_reg));
            scheduler_next_task = fsl_source_reg;
            break;
        case TRC_TASK_END:
            fprintf(trace, "TASK_END %s pc %x\n", 
                    Lookup_Name(fsl_source_reg), fsl_source_reg);
            assert((scheduler_end_task == 0)
                    || (scheduler_end_task == fsl_source_reg));
            scheduler_end_task = fsl_source_reg;
            break;
        case TRC_IN_RTI:
            fprintf(trace, "RETI %s pc %x\n", 
                    Lookup_Name(fsl_source_reg), fsl_source_reg);
            assert((scheduler_end_task == 0)
                    || (scheduler_end_task == fsl_source_reg));
            scheduler_end_task = fsl_source_reg;
            interrupt_handler = 1;
            break;
        default :
            fprintf(stderr, "Unknown capture event %u.\n", data);
            exit(1);
            break;
        }
        C_Report();
        break;
    case 0x51:
        /* notify task name */
        {
            Known_Task * t = Account_Task(fsl_source_reg, 0, 0);
            unsigned i = 0;
            byte = mb_orig.load_fn(m_user, data + i, 1);
            while (byte && (i < MAX_NAME_LEN)) {
                if (!(isprint(byte) || isspace(byte))) {
                    fprintf(stderr, "task name contains unprintable "
                        "byte %d from %x\n", byte, data);
                    exit(1);
                }
                t->name[i] = byte;
                i++;
                byte = mb_orig.load_fn(m_user, data + i, 1);
            } 
            t->name[i] = '\0';
        }
        break;
    default:
        fprintf(stderr, "Unknown FSL put operation %u.\n", fsl);
        exit(1);
        break;
    }
    if (In_User()) {
        user_metric[METRIC_MISS] = miss_count;
    } else {
        sys_metric[METRIC_MISS] = miss_count;
    }
    mb->clock_cycle = clock_cycle;
}

static unsigned C_Get(void * m_user, unsigned flags, unsigned fsl)
{
    fprintf(stderr, "Nothing to get?\n");
    exit(1);
    return 0;
}

static void State_User_to_System(void)
{
    unsigned i;

    fprintf(trace, "state: user -> system\n");
    assert(!In_User());
    assert(num_active_tasks > 0);

    for (i = 0; i < NUM_METRICS; i++) {
        active_task[num_active_tasks - 1].metric[i] += user_metric[i];
        user_metric[i] = 0;
    }
}

static void End_Task(int aborted)
{
    assert(num_active_tasks > 0);
    num_active_tasks--;
    fprintf(trace, "ENDTASK %s pc %x time %ju\n", 
            Lookup_Name(active_task[num_active_tasks].function),
            active_task[num_active_tasks].function,
            active_task[num_active_tasks].metric[METRIC_TIME]);
    Account_Task(active_task[num_active_tasks].function,
            active_task[num_active_tasks].metric, !aborted);
}

static void State_System_to_User(void)
{
    unsigned i;

    fprintf(trace, "state: system -> user\n");
    //assert(In_User());
    if (interrupt_handler || scheduler_next_task) {

        /* was in the ISR */
        if (scheduler_next_task) {
            /* and scheduled a new task! */
            assert(num_active_tasks < MAX_THREADS);
            assert(num_active_tasks >= 1);
            active_task[num_active_tasks-1].metric[METRIC_PREEMPTS]++;

            memset(&active_task[num_active_tasks], 0, sizeof(Active_Task));
            active_task[num_active_tasks].function = scheduler_next_task;
            active_task[num_active_tasks].name = 
                        Lookup_Name(scheduler_next_task);
            num_active_tasks++;
            fprintf(trace, "NEWTASK %s time %ju\n", 
                    active_task[num_active_tasks-1].name, 
                    sys_metric[METRIC_TIME]);
            Account_Task(SPECIAL_SCHED_TO, sys_metric, 1);
        } else {
            fprintf(trace, "SAMETASK %s time %ju\n", 
                    active_task[num_active_tasks-1].name, 
                    sys_metric[METRIC_TIME]);
            Account_Task(SPECIAL_SCHED_FROM, sys_metric, 1);
        }

        if (scheduler_end_task) {
            /* and ended a task */
            assert(num_active_tasks > 0);
            assert(active_task[num_active_tasks-1].function ==
                    scheduler_end_task);
            End_Task(0);
        }
    } else {
        /* was just in some code called by the user */
        for (i = 0; i < NUM_METRICS; i++) {
            user_metric[i] += sys_metric[i];
        }
    }
    scheduler_end_task = scheduler_next_task = 0;
    memset(sys_metric, 0, sizeof(sys_metric));
}

static void State_Exit(void)
{
    unsigned i;
    if (!In_User()) {
        /*State_System_to_User();*/
    }
    if (num_active_tasks) {
        for (i = 0; i < NUM_METRICS; i++) {
            active_task[num_active_tasks - 1].metric[i] += 
                    sys_metric[i] + user_metric[i];
            user_metric[i] = 0;
            sys_metric[i] = 0;
        }
        while(num_active_tasks > 0) {
            End_Task(1);
        }
    }
}

int main ( int argc , char ** argv )
{
    struct MB_System_Context_struct * sys ;
    const char * out ;
    unsigned i, j;
    unsigned rtid = 0;
    unsigned was_in_system, now_in_system;
    uint64_t report_time = 0;

    if (argc < 4) {
        printf("Usage: %s <elf binary> "
                "<trace output.txt> <run budget> [flags]\n", 
                argv[0]);
        return 1 ;
    }
    run_budget = (uint64_t) strtoull(argv[3], NULL, 0);
    if (run_budget == 0) {
        printf("run budget must be non-zero\n");
        return 1;
    }
    trace = fopen(argv[2], "wt");
    if (!trace) {
        printf("Can't create trace file.\n");
        return 1;
    }
    if (argc > 4) {
        dcache_writeback = 0;
        for (i = 0; argv[4][i] != '\0'; i++) {
            switch(argv[4][i]) {
            case 'w':
                dcache_writeback = 1;
                break;
            case 'x':
                dcache_hash = 1;
                break;
            }
        }
    }

    memset(icache_tags, 0xff, sizeof(icache_tags));
    memset(dcache_tags, 0xff, sizeof(dcache_tags));
    memset(dcache_dirty, 0x00, sizeof(dcache_dirty));

    sys = MB_System_Init(Trace, Transfer_Cost(4), NULL);
    mb = MB_System_Get_MB_Context(sys);
    out = MB_Read_Elf(sys, argv[1]);
    if (out != NULL) {
        printf("ELF read error: %s\n", out);
        return 1;
    }
    /* Register devices */
    MB_Map_IO ( sys , TIMER_REGISTER >> PAGE_SHIFT , 
                    Timer_Load , Timer_Store , NULL ) ;
    MB_Map_IO ( sys , BASE_REGISTER_AREA >> PAGE_SHIFT , 
                    Basereg_Load , Basereg_Store , NULL ) ;

    /* setup carousel */
    memcpy(&mb_orig, mb, sizeof(MB_Context));
    mb->ifetch_fn = C_IFetch;
    mb->load_fn = C_Load;
    mb->store_fn = C_Store;
    mb->put_fn = C_Put;
    mb->get_fn = C_Get;

    {
        struct sigaction sa;
        /* SIGABRT is raised by assertion failure */
        sa.sa_sigaction = Program_Aborted;
        sa.sa_flags = SA_SIGINFO;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGABRT, &sa, NULL);

        sa.sa_sigaction = Clean_Stop;
        sa.sa_flags = SA_SIGINFO;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGINT, &sa, NULL);
    }

    /* special tasks */
    num_active_tasks = 1;
    active_task[0].function = SPECIAL_IDLE;
    active_task[0].name = "idle";
    num_known_tasks = 4;
    known_task[0].function = SPECIAL_BOOT;
    known_task[1].function = SPECIAL_IDLE;
    known_task[2].function = SPECIAL_SCHED_TO;
    known_task[3].function = SPECIAL_SCHED_FROM;
    strncpy(known_task[0].name, "boot", MAX_NAME_LEN);
    strncpy(known_task[1].name, "idle", MAX_NAME_LEN);
    strncpy(known_task[2].name, "sys", MAX_NAME_LEN);
    strncpy(known_task[3].name, "sys", MAX_NAME_LEN);
    was_in_system = 1;
    rtid = 100;

    /* Simulate */
    while (stop == HAMMER_TIME) {
        Time_Add(mb->clock_cycle, "iexec");
        mb->clock_cycle = 0;
        rtid++;

        if (mb->cur_iword == 0xb62e0000) { /* RTID */
            rtid = 0;
        }
#if 0
        switch (Get_Mode()) {
        case MODE_SYSTEM:   fprintf(trace, "system\n"); break;
        case MODE_USER:     fprintf(trace, "user\n"); break;
        case MODE_SYSCALL:  fprintf(trace, "syscall\n"); break;
        default:            break;
        }
#endif
        now_in_system = (!In_User()) && (rtid != 1);
        if (was_in_system && !now_in_system) {
            /* system -> user transition */
            State_System_to_User();
        } else if (now_in_system && !was_in_system) {
            /* user -> system transition */
            State_User_to_System();
            if (abs_time >= (run_budget + boot_time)) {
                stop = DUE_TO_END;
            }
        }
        if (mb->msr & MSR_IE) {
            last_ie_time = abs_time;
        } else if (last_ie_time) {
            if ((abs_time - last_ie_time) > longest_cs_time) {
                longest_cs_time = abs_time - last_ie_time;
            }
        }

        MB_Step ( mb , (trigger_time > 0) && (abs_time >= trigger_time));
        was_in_system = now_in_system;
    }

    Flush_Store_Queue();
    Time_Add(mb->clock_cycle, "cleanup");

    State_Exit();
    printf("\n");
    printf("======================\n");
    fflush(trace);

    switch(stop) {
    case DUE_TO_EXIT:       printf("Stop due to exit(%u)\n", mb->gpr[5]); 
                            fprintf(trace, "exit(%u);\n", mb->gpr[5]);
                            return mb->gpr[5];
    case DUE_TO_INTERRUPT:  printf("Stop due to ^C\n");
                            break;
    case DUE_TO_END:        printf("Stop normally\n");
                            break;
    default:                printf("Unknown stop condition\n");
                            break;
    }

    for (i = 0; i < num_known_tasks; i++) {
        const char const * sys = "\nSYS %-8s";
        const char const * usr = "\nUSR %-8s";

        switch(known_task[i].function) {
        case SPECIAL_IDLE:  
            printf(sys, "idle");
            break;
        case SPECIAL_SCHED_TO:  
            printf(sys, "sch to");
            break;
        case SPECIAL_SCHED_FROM:  
            printf(sys, "sch from");
            break;
        case SPECIAL_BOOT:  
            printf(sys, "boot");
            break;
        default:
            printf(usr, known_task[i].name);
            break;
        }
        report_time += known_task[i].total_time;
        printf(" %6u %6u %12ju",
                known_task[i].start_count,
                known_task[i].complete_count,
                known_task[i].total_time);

        for (j = 0; j < NUM_METRICS; j++) {
            if (j == 0) {
                printf("%5s ", metric_name[j]);
            } else {
                printf("%44s ", metric_name[j]);
            }
            printf("%10ju %10ju %10ju\n",
                    known_task[i].metric[j].best,
                    (!known_task[i].complete_count) ? 0 : (
                        known_task[i].metric[j].total /
                            (uint64_t) known_task[i].complete_count),
                    known_task[i].metric[j].worst);
        }
    }
    fclose(trace);

    printf("\ntotal time %ju\n", (abs_time - boot_time));
    printf("CS time %ju\n", longest_cs_time);

    if (report_time != abs_time) {
        fprintf(stderr, "reported time and absolute time do not match! "
            "ticks have been lost or gained!\n"
            "report time = %ju\n"
            "abs time = %ju\n"
            "sys_metric = %ju\n"
            "user_metric = %ju\n", 
            report_time, abs_time,
            sys_metric[METRIC_TIME],
            user_metric[METRIC_TIME]);
        return 1;
    }


    return 0 ;
}

